# Generic makefile for a library
# Last edited on 2011-05-09 13:19:03 by stolfi

SHELL := /bin/bash

######################################################################
# To be included by the top-level Makefile in the library's directory.
# The caller must define
#  
#   ${LIBNAME}     name of library (WITHOUT the directory or the ".a")
#   ${MY_INC_DIR}  directory to search for external ".h" files.
#
# and may also define
#  
#   ${IGNORE} a list of ".h" or ".c" files that should not be compiled or installed.
#   ${OTHER_I_FLAGS}  additional -I flags for C compilation and dependency checks.
#   ${EXTRA_C_FLAGS} additional flags to compile C source files.

ifneq "/${LIBNAME}" "/"

# ----------------------------------------------------------------------
# Package components and directories

# Library file name:
LIBFILE = ${LIBNAME}.a

# Get list of sources:
HFILES := ${filter-out ${IGNORE}, ${wildcard *.h}}
CFILES := ${filter-out ${IGNORE}, ${wildcard *.c}}

# Derived object files:
HOFILES := ${subst .h,.ho,${HFILES}}
OFILES := ${subst .c,.o,${CFILES}}

# ----------------------------------------------------------------------
# Locations of imported packages

# Location of X11 headers and libraries (including ICE):
X11_INC_DIR := /usr/include/X11

# ----------------------------------------------------------------------
# Optional include directories

# Get OpenGL libraries is requested:
ifeq "/${USE_GL}" "/YES" 
  GL_I_FLAGS := \
    -I ${MY_INC_DIR|/GL \
    -I /usr/include/GL
else
  GL_I_FLAGS := 
endif

# Get X11 libraries and headers if requested:
ifeq "/${USE_X11}" "/YES" 
  X11_I_FLAGS := -I${X11_INC_DIR}
else
  X11_I_FLAGS := 
endif
  
#---------------------------------------------------------------------
# Phony "make" targets:

.PHONY:: \
  depend depend-lib \
  all actions \
  build build-lib do-build-lib \
  check \
  clean clean-lib

#---------------------------------------------------------------------
# Default target: 

all: build-lib

#---------------------------------------------------------------------
# "make actions" performs the actions listed in ${ACTIONS}, if any

actions:
ifeq "/${ACTIONS}" "/"
	@echo '$${ACTIONS} is empty'
else
	${MAKE} ${ACTIONS}
endif

#---------------------------------------------------------------------
# Directories to search for #include files:

I_FLAGS := \
  -I. \
  -I${MY_INC_DIR} \
  ${GL_I_FLAGS} \
  ${X11_I_FLAGS} \
  ${OTHER_I_FLAGS}

#---------------------------------------------------------------------
# "make depend" recreates the source dependency file ${DEPFILE}.

DEPFILE := Deps.make

depend: depend-lib

depend-lib: ${DEPFILE}

${DEPFILE}: ../extract-ho-deps.sh ${HFILES} ${CFILES}
	@/bin/rm -f ${DEPFILE}
	@../extract-ho-deps.sh ${I_FLAGS} -I/usr/include \
            ${HFILES} ${CFILES} \
          | egrep -v ': /usr/include' \
          > ${DEPFILE}

#---------------------------------------------------------------------
# "make build" and "make build-lib" make sure that the ${DEPFILE}
# is up to date, and then recreate the object files from the source files
# as necessary.  "make build" and "make build-tests" also rebuild
# the test programs.

build: build-lib build-tests
  
build-lib: ${DEPFILE}
	${MAKE} do-build-lib
        
do-build-lib: ${LIBFILE}

CC := /usr/bin/gcc

C_FLAGS := \
  -Wall -Wundef \
  -Wpointer-arith \
  -Wmissing-prototypes \
  -ggdb \
  -fpcc-struct-return \
  -ffloat-store \
  -frounding-math

%.o: %.c
	${CC} -c ${C_FLAGS} ${EXTRA_C_FLAGS} ${I_FLAGS} $*.c
  
%.ho: %.h
	${CC} -o $*.ho -c ${C_FLAGS} ${EXTRA_C_FLAGS} ${I_FLAGS} -x c $*.h \
          || /bin/rm -f $*.ho
  
${LIBFILE}: ${HOFILES} ${OFILES}
	-rm -f $*.a
	ar crv $*.a ${OFILES}
	ranlib $*.a

ifneq "/${wildcard ${DEPFILE}}" "/"
  # Include specific dependencies extracted by "make depend"
  include ${DEPFILE}
endif

#---------------------------------------------------------------------
# "make clean" deletes all derived files.

clean: clean-lib

clean-lib:
	-/bin/rm -f *.o *.ho *.a core
	touch --date='Jan 01 1970' ${PROG}.deps
	${MAKE} depend
  
endif
# End of ${LIBNAME} section.
######################################################################
